#define G_LOG_DOMAIN "boilerplate-window"

#include "config.h"

#include <gtk/gtk.h>

#include "boilerplate-application.h"
#include "boilerplate-window.h"

typedef struct
{
    GtkWidget *toolbar;
    GtkWidget *headerbar;
    GtkWidget *content;
} BoilerplateWindowPrivate;

G_DEFINE_TYPE_WITH_PRIVATE (BoilerplateWindow, boilerplate_window, GTK_TYPE_APPLICATION_WINDOW)

//TODO: Define docs for new functions

// https://developer.gnome.org/gtk3/stable/GtkWidget.html#GtkWidget-size-allocate
static void
boilerplate_window_size_allocate (GtkWidget    *widget,
                                  GdkRectangle *allocation)
{
    BoilerplateWindow *self = BOILERPLATE_WINDOW (widget);
    int width,
        height,
        x,
        y;

    gtk_window_get_size (GTK_WINDOW (widget),
                         &width,
                         &height);

    gtk_window_get_position (GTK_WINDOW (widget),
                             &x,
                             &y);

    if (!(self->state->maximized || self->state->fullscreen)) {
        g_object_set (self->state,
                      "width", width,
                      "height", height,
                      "x", x,
                      "y", y,
                      NULL);
    }
}

// https://developer.gnome.org/gtk3/stable/GtkWidget.html#GtkWidget-window-state-event
// https://wiki.gnome.org/HowDoI/SaveWindowState
// https://developer.gnome.org/gtk3/stable/GtkWidget.html#GtkWidget-window-state-event
// http://git.mate-desktop.org/atril/plain/shell/ev-window.c
static gboolean
boilerplate_window_window_state_event (GtkWidget           *widget,
                                       GdkEventWindowState *event)
{
    BoilerplateWindow *self = BOILERPLATE_WINDOW (widget);
    gboolean state_changed = GDK_EVENT_PROPAGATE;
    bool maximized;
    bool fullscreen;

    if (GTK_WIDGET_CLASS (boilerplate_window_parent_class)->window_state_event != NULL) {
        state_changed = GTK_WIDGET_CLASS (boilerplate_window_parent_class)->window_state_event (widget, event);
    }

    maximized  = (event->new_window_state & GDK_WINDOW_STATE_MAXIMIZED) ? 1 : 0; // != 0
    fullscreen = (event->new_window_state & GDK_WINDOW_STATE_FULLSCREEN) ? 1 : 0;

    g_object_set (self->state,
                  "maximized", maximized,
                  "fullscreen", fullscreen,
                  NULL);

    //TODO: Return more then false
    return state_changed;
}

static void
boilerplate_window_destroy (GtkWidget *widget)
{
    BoilerplateWindow *self = BOILERPLATE_WINDOW (widget);

    GTK_WIDGET_CLASS (boilerplate_window_parent_class)->destroy (widget);
}

static void
boilerplate_window_constructed (GObject *object)
{
    BoilerplateWindow *self = BOILERPLATE_WINDOW (object);

    //TODO: Need a gobject type!
    // https://github.com/GNOME/gtk/tree/gtk-3-22/demos/gtk-demo
    self->state = g_object_new (window_state_get_type (), NULL);

    GSettings *settings = g_settings_new (PACKAGE_ID ".window-state");

    g_settings_bind (settings, "width",
                     self->state, "width",
                     G_SETTINGS_BIND_DEFAULT);
    g_settings_bind (settings, "height",
                     self->state, "height",
                     G_SETTINGS_BIND_DEFAULT);
    g_settings_bind (settings, "x",
                     self->state, "x",
                     G_SETTINGS_BIND_DEFAULT);
    g_settings_bind (settings, "y",
                     self->state, "y",
                     G_SETTINGS_BIND_DEFAULT);
    g_settings_bind (settings, "maximized",
                     self->state, "maximized",
                     G_SETTINGS_BIND_DEFAULT);
    g_settings_bind (settings, "fullscreen",
                     self->state, "fullscreen",
                     G_SETTINGS_BIND_DEFAULT);

    gtk_window_set_default_size (GTK_WINDOW (self),
                                 self->state->width,
                                 self->state->height);

    if (self->state->maximized)
        gtk_window_maximize (GTK_WINDOW (self));

    if (self->state->fullscreen)
        gtk_window_fullscreen (GTK_WINDOW (self));

    G_OBJECT_CLASS (boilerplate_window_parent_class)->constructed (object);
}

/**
 * boilerplate_window_finalize:
 * @object: The window
 *
 * Does any last minute clean-ups when the class ends
 *
 * Returns: void
 *
 * Since: 0.1.0
 */
static void
boilerplate_window_finalize (GObject *object)
{
    BoilerplateWindow *self = BOILERPLATE_WINDOW (object);
    BoilerplateWindowPrivate* priv = boilerplate_window_get_instance_private (self);

    g_object_unref (priv->toolbar);

    G_OBJECT_CLASS (boilerplate_window_parent_class)->finalize(object);
}

/**
 * boilerplate_window_init:
 * @self: The window
 *
 * Initializes the application window.
 *
 * Returns: void
 *
 * Since: 0.1.0
 */
static void
boilerplate_window_init (BoilerplateWindow *self)
{
    BoilerplateWindowPrivate* priv = boilerplate_window_get_instance_private (self);
    GtkBuilder *builder;

    /* Init window.ui template */
    gtk_widget_init_template (GTK_WIDGET (self));

    /* Set window title */
    gtk_window_set_title (GTK_WINDOW (self), PACKAGE_NAME);

    /* Adding headerbar/titlebar/toolbar to private struct */
    if (priv->toolbar == NULL) {
        builder = gtk_builder_new_from_resource (PACKAGE_RESOURCES "/ui/toolbar.ui");
        priv->toolbar = GTK_WIDGET (gtk_builder_get_object (builder, "toolbar"));

        /* Ref toolbar so we can move it between the headerbar and toolbar */
        g_object_ref (priv->toolbar);

        g_object_unref (builder);
    }

    /* Set default window decorations (a.k.a GtkHeaderBar) */
    gtk_window_set_titlebar (GTK_WINDOW (self), priv->headerbar);
    gtk_container_add (GTK_CONTAINER (priv->headerbar), priv->toolbar);
}

/**
 * boilerplate_window_class_init:
 * @class: The window class
 *
 * Initializes the various functions in the window and links the gtk_widget
 * template loaded from gresource.
 *
 * Returns: void
 *
 * Since: 0.1.0
 */
static void
boilerplate_window_class_init (BoilerplateWindowClass *klass)
{
    GObjectClass *object_class = G_OBJECT_CLASS (klass);
    GtkWidgetClass *widget_class = GTK_WIDGET_CLASS (klass);

    object_class->finalize = boilerplate_window_finalize;
    object_class->constructed = boilerplate_window_constructed;

    widget_class->destroy = boilerplate_window_destroy;
    widget_class->window_state_event = boilerplate_window_window_state_event;
    widget_class->size_allocate = boilerplate_window_size_allocate;

    gtk_widget_class_set_template_from_resource (widget_class,
                                                 PACKAGE_RESOURCES "/ui/window.ui");

    gtk_widget_class_bind_template_child_private (widget_class, BoilerplateWindow, headerbar);
    gtk_widget_class_bind_template_child_private (widget_class, BoilerplateWindow, content);
}

/**
 * boilerplate_window_new:
 * @application: The application
 *
 * Creates a new GtkApplicationWindow.
 *
 * Returns: BoilerplateWindow -> GtkApplicationWindow
 *
 * Since: 0.1.0
 */
BoilerplateWindow *
boilerplate_window_new (BoilerplateApplication *application)
{
    return g_object_new (BOILERPLATE_TYPE_WINDOW,
                         "application", application,
                         "title", PACKAGE_NAME,
                         NULL);
}

